;;;; error-utils.scm
;;;; Kon Lovett, Aug '10
;;;; Kon Lovett, may '17

(module error-utils

(;export
  ;
  error/no-raise *error/no-raise
  ;
  errorf errorf/no-raise
  ;
  error-format-procedure)

(import scheme)

(import chicken)
(import (only extras format) )
(import (only data-structures chop ->string string-intersperse) )

(import (only (srfi 1) append!) )
(require-library (srfi 1))

;;;

;; Print error message but don't throw an exception
;;
(define (error/no-raise . args)
  (receive (port args) (error-port-args args)
    (*error/no-raise args port) ) )

;; Print error-style message to port
;; unlike 'error' will print arguments when loc but no msg
;;
(define (*error/no-raise args #!optional (port (current-error-port)))
  (receive (loc msg args) (error-params args)
    (let (
        (errmsg
          (string-append
            "\n"
            "Error" (if (or loc msg) ": " "")
            (if (and loc msg) (string-append "(" (->string loc) ")" " ") "")
            (or (and msg (->string msg)) (and loc (->string loc)) "")
            (cond
              ((null? args)
                "" )
              ((null? (cdr args))
                (string-append ": " (->sexpr-string (car args))) )
              (else
                ;leading empty string so leading newline
                (string-intersperse (append '("") (map ->sexpr-string args)) "\n") ) )
            "\n" )))
      (display errmsg port) ) ) )

;;
(define (errorf . args)
  (*errorf #f error args) )

;;
(define (errorf/no-raise . args)
  (receive (port args) (error-port-args args)
    (*errorf port error/no-raise args) ) )

;; Format procedure for error
;;
(define error-format-procedure
  (make-parameter
    format
    (lambda (x)
      (if (procedure? x)
        x
        (begin
          (warning 'error-format-procedure "invalid procedure" x)
          (error-format-procedure) ) ) ) ) )

;;;

;;
(define (*errorf port proc args)
  (let ((portarg (if port `(,port) '())))
    (receive (loc fmt fmtargs) (error-params args)
      (let (
          (err-args
            (if (not fmt)
              (append args portarg)
              (let ((msg (apply (error-format-procedure) #f fmt fmtargs)))
                (if loc
                  (append! `(,loc ,msg) portarg)
                  (append! `(,msg) portarg) ) ) ) ) )
        (apply proc err-args) ) ) ) )

;;
(define (->sexpr-string obj)
  ((error-format-procedure) #f "~S" obj) )

;; Parse error style argument list into (values loc msg args)
;;
(define (error-params args)
  (let* (
      (loc
        (and
          (not (null? args))
          (symbol? (car args)) (car args)) )
      (msg
        (if (not loc)
          (and
            (not (null? args))
            (string? (car args)) (car args))
          (and
            (not (null? (cdr args)))
            (string? (cadr args)) (cadr args)) ) )
      (args
        (if (and loc msg)
          (cddr args)
          (if (or loc msg)
            (cdr args)
            args ) ) ) )
    (values loc msg args) ) )

;; Parse error/no-raise style argument list into (values port args)
;;
(define (error-port-args args)
  (let* ((len (length args) )
         (argls (and (<= 2 len) (chop args (fx- len 1))) )
         (args (if argls (car argls) args) )
         (port (and argls (caadr argls)) )
         (args (if (port? port) args (append args (if port `(,port) '()))) )
         (port (if (port? port) port (current-error-port)) ) )
    (values port args) ) )

) ;module error-utils
